"""A research library for multi-agent learning."""


# Import all submodules.
from coaction import agents
from coaction import experiments
from coaction import games
from coaction import loggers
from coaction import utils


__all__ = ["agents", "experiments", "games", "loggers", "utils"]
